@@ -6,6 +6,9 @@
 /* The NOFFLE config-file */
 #undef CONFIGFILE
 
+/* The NOFFLE users-file */
+#undef USERSFILE
+
 /* The default NOFFLE spool-directory */
 #undef SPOOLDIR
 
@@ -18,6 +21,12 @@
 /* path to sendmail binary */
 #undef SENDMAILPROG
 
+/* Is authentication enabled? */
+#undef USE_AUTH
+
+/* Are we using PAM for authentication? */
+#undef USE_PAM
+
 @TOP@
 /* autoheader generated things inserted here. */
 @BOTTOM@
