@@ -1,15 +1,13 @@
 /*
   client.c
 
-  $Id: client.c,v 1.9.2.11 2001/05/14 07:08:39 enz Exp $
+  $Id: client.c,v 1.44 2003/06/05 17:55:44 godisch Exp $
 */
 
 #if HAVE_CONFIG_H
 #include <config.h>
 #endif
 
-#include "client.h"
-
 #include <stdio.h>
 #include <arpa/inet.h>
 #include <ctype.h>
@@ -19,11 +17,15 @@
 #include <stdarg.h>
 #include <sys/socket.h>
 #include <unistd.h>
+#include "client.h"
 #include "configfile.h"
 #include "content.h"
+#include "control.h"
 #include "dynamicstring.h"
+#include "filter.h"
 #include "group.h"
 #include "itemlist.h"
+#include "lock.h"
 #include "log.h"
 #include "over.h"
 #include "protocol.h"
@@ -33,32 +35,6 @@
 #include "wildmat.h"
 #include "portable.h"
 
-/*
-  Some newsgroups names are reserved for server-specific or server
-  pseudo groups. We don't want to fetch them. For example, INN
-  keeps all its control messages in a 'control' hierarchy, and
-  used the "to." hierarchy for dark and mysterious purposes I think
-  are to do with newsfeeds. The recommended restrictions are documented
-  in C.Lindsay, "News Article Format", <draft-ietf-usefor-article-03.txt>.
-*/
-
-struct ForbiddenGroupName
-{
-    const char *pattern;
-    Bool match;
-} forbiddenGroupNames[] =
-{
-    { "*.*", FALSE },			/* Single component */
-    { "control.*", TRUE },		/* control.* groups */
-    { "to.*", TRUE },			/* to.* groups */
-    { "*.all", TRUE },			/* 'all' as a component */
-    { "*.all.*", TRUE },
-    { "all.*", TRUE },
-    { "*.ctl", TRUE },			/* 'ctl' as a component */
-    { "*.ctl.*", TRUE },
-    { "ctl.*", TRUE }
-};
-
 struct
 {
     FILE* in;       /* Receiving socket from server */
@@ -69,16 +45,14 @@ struct
     int rmtFirst;   /* First article of current group at server */
     int rmtLast;    /* Last article of current group at server */
     Bool auth;      /* Authentication already done? */
-    Bool connected; /* Connection alive? */
     Str serv;       /* Remote server name */
-} client = { NULL, NULL, "", "", "", 1, 0, FALSE, FALSE, "" };
+} client = { NULL, NULL, "", "", "", 1, 0, FALSE, "" };
 
 static void
-breakDown( void )
+logBreakDown( void )
 {
     Log_err( "Connection to remote server lost "
              "(article numbers could be inconsistent)" );
-    client.connected = FALSE;
 }
 
 static Bool
@@ -86,11 +60,9 @@ getLn( Str line )
 {
     Bool r;
 
-    if ( ! client.connected )
-        return FALSE;
     r = Prt_getLn( line, client.in, Cfg_connectTimeout() );
     if ( ! r )
-        breakDown();
+        logBreakDown();
     return r; 
 }
 
@@ -99,11 +71,9 @@ getTxtLn( Str line, Bool *err )
 {
     Bool r;
 
-    if ( ! client.connected )
-        return FALSE;
     r = Prt_getTxtLn( line, err, client.in, Cfg_connectTimeout() );
     if ( *err )
-        breakDown();
+        logBreakDown();
     return r; 
 }
 
@@ -112,7 +82,7 @@ putTxtBuf( const char *buf )
 {
     Prt_putTxtBuf( buf, client.out );
     fflush( client.out );
-    Log_dbg( "[S FLUSH]" );
+    Log_dbg( LOG_DBG_PROTOCOL, "[S FLUSH]" );
 }
 
 static void
@@ -120,7 +90,7 @@ putEndOfTxt( void )
 {
     Prt_putEndOfTxt( client.out );
     fflush( client.out );
-    Log_dbg( "[S FLUSH]" );
+    Log_dbg( LOG_DBG_PROTOCOL, "[S FLUSH]" );
 }
 
 static Bool
@@ -129,15 +99,13 @@ putCmdLn( const char *line )
     Bool err;
     unsigned int n;
 
-    if ( ! client.connected )
-        return FALSE;
-    strcpy( client.lastCmd, line );
-    strcpy( client.lastStat, "[no status available]" );
-    Log_dbg( "[S] %s", line );
+    Utl_cpyStr( client.lastCmd, line );
+    Utl_cpyStr( client.lastStat, "[no status available]" );
+    Log_dbg( LOG_DBG_PROTOCOL, "[S] %s", line );
     n = fprintf( client.out, "%s\r\n", line );
     err = ( n != strlen( line ) + 2 );
     if ( err )
-        breakDown();
+        logBreakDown();;
     return ! err;
 }
 
@@ -153,7 +121,7 @@ putCmd( const char *fmt, ... )
     if ( ! putCmdLn( line ) )
         return FALSE;
     fflush( client.out );
-    Log_dbg( "[S FLUSH]" );
+    Log_dbg( LOG_DBG_PROTOCOL, "[S FLUSH]" );
     return TRUE;
 }
 
@@ -171,7 +139,7 @@ putCmdNoFlush( const char *fmt, ... )
 
 static int getStat( void );
 
-static Bool
+static int
 performAuth( void )
 {
     int stat;
@@ -181,30 +149,30 @@ performAuth( void )
     if ( strcmp( user, "" ) == 0 )
     {
         Log_err( "No username for authentication set" );
-        return FALSE;
+        return STAT_AUTH_REQUIRED;
     }    
     putCmd( "AUTHINFO USER %s", user );
     stat = getStat();
-    if ( stat == STAT_AUTH_ACCEPTED )
-        return TRUE;
-    else if ( stat != STAT_MORE_AUTH_REQUIRED )
+    if ( stat == STAT_AUTH_ACCEPTED
+         || stat == STAT_AUTH_ACCEPTED_DEPREC )
+        return stat;
+    else if ( stat != STAT_MORE_AUTH_REQUIRED_DEPREC
+              && stat != STAT_MORE_AUTH_REQUIRED )
     {
         Log_err( "Username rejected. Server stat: %s", client.lastStat );
-        return FALSE;
+        return stat;
     }    
     if ( strcmp( pass, "" ) == 0 )
     {
         Log_err( "No password for authentication set" );
-        return FALSE;
+        return STAT_AUTH_REQUIRED;
     }
     putCmd( "AUTHINFO PASS %s", pass );
     stat = getStat();
-    if ( stat != STAT_AUTH_ACCEPTED )
-    {
+    if ( stat != STAT_AUTH_ACCEPTED
+         && stat != STAT_AUTH_ACCEPTED_DEPREC)
         Log_err( "Password rejected. Server status: %s", client.lastStat );
-        return FALSE;
-    }    
-    return TRUE;    
+    return stat;    
 }
 
 static int
@@ -214,17 +182,21 @@ getStat( void )
     Str lastCmd;
 
     if ( ! getLn( client.lastStat ) )
-        result = STAT_PROGRAM_FAULT;
+        result = STAT_CONNECTION_LOST;
     else if ( sscanf( client.lastStat, "%d", &result ) != 1 )
     {
         Log_err( "Invalid server status: %s", client.lastStat );
         result = STAT_PROGRAM_FAULT;
     }
-    if ( result == STAT_AUTH_REQUIRED && ! client.auth )
+    if ( ( result == STAT_AUTH_REQUIRED
+           || result == STAT_AUTH_REQUIRED_DEPREC )
+         && ! client.auth )
     {
         client.auth = TRUE;
-        strcpy( lastCmd, client.lastCmd );
-        if ( performAuth() )
+        Utl_cpyStr( lastCmd, client.lastCmd );
+	result = performAuth();
+        if ( result == STAT_AUTH_ACCEPTED
+             || result == STAT_AUTH_ACCEPTED_DEPREC )
         {
             putCmd( lastCmd );
             return getStat();
@@ -241,32 +213,15 @@ connectAlarm( int sig )
     return;
 }
 
-static sig_t
-installSignalHandler( int sig, sig_t handler )
-{
-    struct sigaction act, oldAct;
-
-    act.sa_handler = handler;
-    sigemptyset( &act.sa_mask );
-    act.sa_flags = 0;
-    if ( sig == SIGALRM )
-        act.sa_flags |= SA_INTERRUPT;
-    else
-        act.sa_flags |= SA_RESTART;
-    if ( sigaction( sig, &act, &oldAct ) < 0 )
-        return SIG_ERR;
-    return oldAct.sa_handler;
-}
-
 static Bool
 connectWithTimeout( int sock, const struct sockaddr *servAddr,
                     socklen_t addrLen )
 {
-    sig_t oldHandler;
+    SignalHandler oldHandler;
     int r, to;
 
-    oldHandler = installSignalHandler( SIGALRM, connectAlarm );
-    if ( oldHandler == SIG_ERR )
+    oldHandler = Utl_installSignalHandler( SIGALRM, connectAlarm );
+    if ( oldHandler == (SignalHandler) SIG_ERR )
     {
         Log_err( "client.c:connectWithTimeout: signal failed." );
         return FALSE;
@@ -276,43 +231,67 @@ connectWithTimeout( int sock, const struct sockaddr *servAddr,
         Log_err( "client.c:connectWithTimeout: Alarm was already set." );
     r = connect( sock, servAddr, addrLen );
     alarm( 0 );
-    installSignalHandler( SIGALRM, oldHandler );
+    Utl_installSignalHandler( SIGALRM, oldHandler );
     return ( r >= 0 );
 }
 
+static DynStr *
+collectTxt( void )
+{
+    DynStr *res;
+    Str line;
+    Bool err;
+
+    res = new_DynStr( MAXCHAR );
+    if ( res == NULL )
+	return NULL;
+
+    while ( getTxtLn( line, &err ) && ! err )
+	DynStr_appLn( res, line );
+
+    if ( err )
+    {
+	del_DynStr( res );
+	return NULL;
+    }
+    else
+	return res;
+}
+
 Bool
 Client_connect( const char *serv )
 {
-    unsigned short int port;
+    long port;
     int sock, i;
     unsigned int stat;
     struct hostent *hp;
     char *pStart, *pColon;
     Str host, s;
+    Str user, pass;
     struct sockaddr_in sIn;
 
-    client.connected = FALSE;
+    ASSERT( client.in == NULL && client.out == NULL );
     client.auth = FALSE;
     Utl_cpyStr( s, serv );
     pStart = Utl_stripWhiteSpace( s );
     pColon = strstr( pStart, ":" );
     if ( pColon == NULL )
     {
-        strcpy( host, pStart );
+        Utl_cpyStr( host, pStart );
         port = 119;
     }
     else
     {
         *pColon = '\0';
-        strcpy( host, pStart );
-        if ( sscanf( pColon + 1, "%hi", &port ) != 1 )
+        Utl_cpyStr( host, pStart );
+        if ( sscanf( pColon + 1, "%li", &port ) != 1 )
         {
             Log_err( "Syntax error in server name: '%s'", serv );
             return FALSE;;
         }
         if ( port <= 0 || port > 65535 )
         {
-            Log_err( "Invalid port number %hi. Must be in [1, 65535]", port );
+            Log_err( "Invalid port number %hu. Must be in [1, 65535]", port );
             return FALSE;;
         }
     }
@@ -340,23 +319,23 @@ Client_connect( const char *serv )
 		if ( client.out != NULL )
 		    fclose( client.out );
                 close( sock );
+		client.in = client.out = NULL;
                 break;
             }
             Utl_cpyStr( client.serv, serv );
-            client.connected = TRUE;
             stat = getStat();
-	    if ( stat == STAT_READY_POST_ALLOW ||
-		 stat == STAT_READY_NO_POST_ALLOW )
+            switch( stat )
 	    {
-		/* INN needs a MODE READER before it will permit POST. */
-		putCmd( "MODE READER" );
-		stat = getStat();
-	    }
-            switch( stat ) {
             case STAT_READY_POST_ALLOW:
             case STAT_READY_NO_POST_ALLOW: 
                 Log_inf( "Connected to %s:%d",
                          inet_ntoa( sIn.sin_addr ), port );
+		/* INN needs a MODE READER before it will permit POST. */
+		putCmd( "MODE READER" );
+		getStat();
+                Cfg_authInfo( client.serv, user, pass );
+                if ( strcmp( user, "" ) != 0 )
+                    performAuth();
                 return TRUE;
             default:
                 Log_err( "Bad server stat %d", stat ); 
@@ -365,54 +344,91 @@ Client_connect( const char *serv )
 	    fclose( client.in );
 	    fclose( client.out );
 	    close( sock );
+	    client.in = client.out = NULL;
         }
     }
-    client.connected = FALSE;
     return FALSE;
 }
 
 static Bool
-isForbiddenGroupName( const char *name )
+isGetGroup( const char *name )
 {
-    size_t i;
+    GroupEnum *ge;
+    Bool emptyList;
+    const char *pattern;
 
-    for ( i = 0;
-	  i < sizeof( forbiddenGroupNames ) /
-	      sizeof( struct ForbiddenGroupName );
-	  i++ )
+    emptyList = TRUE;
+    ge = new_GetGrEn( client.serv );
+    while ( ( pattern = GrEn_next( ge ) ) != NULL )
     {
-	/* Negate result of Wld_match to ensure it is 1 or 0. */
-	if ( forbiddenGroupNames[i].match !=
-	     ( ! Wld_match( name, forbiddenGroupNames[i].pattern ) ) )
+	emptyList = FALSE;
+	if ( Wld_match( name, pattern ) )
+	{
+	    del_GrEn( ge );
 	    return TRUE;
+	}
     }
-
-    return FALSE;
+    
+    del_GrEn( ge );
+    return emptyList;
 }
 
-static void
-processGrps( void )
+static Bool
+isOmitGroup( const char *name )
+{
+    GroupEnum *ge;
+    const char *pattern;
+
+    ge = new_OmitGrEn( client.serv );
+    while ( ( pattern = GrEn_next( ge ) ) != NULL )
+	if ( Wld_match( name, pattern ) )
+	{
+	    del_GrEn( ge );
+	    return TRUE;
+	}
+    
+    del_GrEn( ge );
+    return FALSE;   
+}
+
+static Bool
+processGrps( Bool noServerPattern )
 {
     char postAllow;
+    Bool groupupdate;
     Bool err;
     int first, last;
-    Str grp, line, file;
-    Bool groupupdate;
-    
+    Str grp, file;
+    Str line;
+
+    ASSERT( ! Lock_gotLock() );
+    if ( ! Lock_openDatabases() )
+	return TRUE;    /* silently ignore */
+
     groupupdate = FALSE;
-    while ( getTxtLn( line, &err ) && ! err )
+
+    while ( getTxtLn( line, &err ) && !err )
     {
-        if ( sscanf( line, "%s %d %d %c",
+        if ( sscanf( line, MAXCHAR_FMT " %d %d %c",
                      grp, &last, &first, &postAllow ) != 4 )
         {
             Log_err( "Unknown reply to LIST or NEWGROUPS: %s", line );
             continue;
         }
-	if ( isForbiddenGroupName( grp ) )
+        if ( ! Grp_isValidName( grp ) )
+        {
+            Log_inf( "Group name %s invalid", grp );
+            continue;
+        }
+	if ( Grp_isForbiddenName( grp ) )
 	{
 	    Log_inf( "Group %s forbidden", grp );
 	    continue;
 	}
+	if ( noServerPattern && ! isGetGroup( grp ) )
+	    continue;
+	if ( isOmitGroup( grp ) )
+	    continue;
         if ( ! Grp_exists( grp ) )
         {
             Log_inf( "Registering new group '%s'", grp );
@@ -423,14 +439,14 @@ processGrps( void )
                 Grp_setFirstLast( grp, first, first - 1 );
             else
                 Grp_setFirstLast( grp, 1, 0 );
-            Grp_setRmtNext( grp, first );
             Grp_setServ( grp, client.serv );
 	    Grp_setPostAllow( grp, postAllow );
 	    groupupdate = TRUE;
         }
         else
         {
-            if ( Cfg_servIsPreferential( client.serv, Grp_server( grp ) ) )
+            if ( ! Grp_local( grp ) && \
+                Cfg_servIsPreferential( client.serv, Grp_server( grp ) ) )
             {
                 Log_inf( "Changing server for '%s': '%s'->'%s'",
                          grp, Grp_server( grp ), client.serv );
@@ -440,23 +456,28 @@ processGrps( void )
 		groupupdate = TRUE;
             }
             else
-                Log_dbg( "Group %s is already fetched from %s",
-                           grp, Grp_server( grp ) );
-            
+                Log_dbg( LOG_DBG_FETCH,
+			 "Group %s is already fetched from %s",
+			 grp, Grp_server( grp ) );            
         }
     }
-    if ( ! err )
+
+    snprintf( file, MAXCHAR, "%s/lastupdate.%s",
+	      Cfg_spoolDir(), client.serv );
+    Utl_stamp( file );
+    if ( groupupdate )
     {
-	snprintf( file, MAXCHAR, "%s/lastupdate.%s",
-		  Cfg_spoolDir(), client.serv );
+	snprintf( file, MAXCHAR, "%s/groupinfo.lastupdate",
+		  Cfg_spoolDir() );
 	Utl_stamp( file );
-	if ( groupupdate )
-	{
-	    snprintf( file, MAXCHAR, "%s/groupinfo.lastupdate",
-		      Cfg_spoolDir() );
-	    Utl_stamp( file );
-	}
     }
+
+    /* I'm absolutely not sure about this. */
+    if ( err && groupupdate )
+            Log_err( "Group list may be corrupted with bogus data." );
+
+    Lock_closeDatabases();
+    return !err;
 }
 
 void
@@ -467,67 +488,186 @@ Client_disconnect( void )
     fclose( client.in );
     fclose( client.out );
     client.in = client.out = NULL;
-    client.connected = FALSE;
 }
 
-Bool
+static int
+doGetGrps( const char *pattern, Bool *noServerPattern )
+{
+    Str cmd;
+    int stat;
+
+    Utl_cpyStr( cmd, "LIST ACTIVE" );
+    if ( pattern[ 0 ] != '\0' )
+    {
+	Utl_catStr( cmd, " " );
+	Utl_catStr( cmd, pattern );
+    }
+
+    *noServerPattern = FALSE;
+    if ( ! putCmd( cmd ) )
+        return STAT_CONNECTION_LOST;
+    stat = getStat();
+    if ( IS_FATAL( stat ) )
+        return stat;
+
+    /*
+     * Try LIST instead of LIST ACTIVE in case server doesn't
+     * support LIST ACTIVE.
+     */
+    if ( stat != STAT_GRPS_FOLLOW )
+    {
+	if ( pattern[ 0 ] != '\0' )
+	    *noServerPattern = TRUE;
+	if ( ! putCmd( "LIST" ) )
+	    return STAT_CONNECTION_LOST;
+	stat = getStat();
+    }    
+    if ( stat != STAT_GRPS_FOLLOW )    
+    {
+	Log_err( "%s failed: %s", cmd, client.lastStat );
+	return stat;
+    }
+
+    
+    if ( processGrps( *noServerPattern ) == FALSE )
+	return STAT_CONNECTION_LOST;
+
+    return STAT_OK;
+}
+
+int
 Client_getGrps( void )
 {
-    if ( ! client.connected )
-        return FALSE;
-    if ( ! putCmd( "LIST" ) )
-        return FALSE;
-    if ( getStat() != STAT_GRPS_FOLLOW )
+    GroupEnum *ge;
+    const char *pattern;
+    Bool doneOne, noServerPattern;
+    int res;
+
+    Log_inf( "Getting groups" );
+
+    doneOne = FALSE;
+    res = STAT_OK;
+    ge = new_GetGrEn( client.serv );
+    while ( res == STAT_OK && ( pattern = GrEn_next( ge ) ) != NULL )
     {
-        Log_err( "LIST command failed: %s", client.lastStat );
-        return FALSE;
+	res = doGetGrps( pattern, &noServerPattern );
+	doneOne = TRUE;
+	if ( noServerPattern )
+	    break;
     }
-    processGrps();
-    return TRUE;
+
+    if ( ! doneOne )
+	res = doGetGrps( "", &noServerPattern );
+
+    del_GrEn( ge );
+    return res;
 }
 
-Bool
-Client_getDsc( void )
+static int
+doGetDsc( const char *pattern, Bool *noServerPattern )
 {
-    Bool err;
-    Str name, line, dsc;
+    Str name, line, dsc, cmd;
+    int stat;
+    DynStr *response;
+    const char *lines;
+    Bool result;
 
-    if ( ! client.connected )
-        return FALSE;
-    Log_inf( "Querying group descriptions" );
-    if ( ! putCmd( "LIST NEWSGROUPS" ) )
-        return FALSE;
-    if ( getStat() != STAT_GRPS_FOLLOW )
+    ASSERT( ! Lock_gotLock() );
+    Utl_cpyStr( cmd, "LIST NEWSGROUPS" );
+    if ( pattern[ 0 ] != '\0' )
     {
-        Log_err( "LIST NEWSGROUPS failed: %s", client.lastStat );
-        return FALSE;
+	Utl_catStr( cmd, " " );
+	Utl_catStr( cmd, pattern );
     }
-    while ( getTxtLn( line, &err ) && ! err )
+
+    *noServerPattern = FALSE;
+    if ( ! putCmd( cmd ) )
+        return STAT_CONNECTION_LOST;
+    stat = getStat();
+    if ( IS_FATAL( stat ) )
+        return stat;
+
+    /* Try without pattern in case server doesn't support patterns. */
+    if ( pattern[ 0 ] != '\0' && stat != STAT_GRPS_FOLLOW )
     {
-        if ( sscanf( line, "%s", name ) != 1 )
+	*noServerPattern = TRUE;
+	if ( !putCmd( "LIST NEWSGROUPS" ) )
+	     return STAT_CONNECTION_LOST;
+	stat = getStat();
+    }
+    if ( stat != STAT_GRPS_FOLLOW )
+    {
+        Log_err( "%s failed: %s", cmd, client.lastStat );
+        return stat;
+    }
+
+    response = collectTxt();
+    if ( response == NULL )
+	return STAT_CONNECTION_LOST;
+    
+    if ( ! Lock_openDatabases() )
+	return STAT_NEWSBASE_FATAL;
+    
+    lines = DynStr_str( response );
+    result = STAT_OK;
+    while ( ( lines = Utl_getLn( line, lines) ) != NULL )
+    {
+        if ( sscanf( line, MAXCHAR_FMT, name ) != 1 )
         {
             Log_err( "Unknown reply to LIST NEWSGROUPS: %s", line );
-            continue;
+	    result = STAT_PROGRAM_FAULT;
+	    break;
         }
-        strcpy( dsc, Utl_restOfLn( line, 1 ) );
+	if ( *noServerPattern && ! isGetGroup( name ) )
+	    continue;
+        Utl_cpyStr( dsc, Utl_restOfLn( line, 1 ) );
         if ( Grp_exists( name ) )
         {
-            Log_dbg( "Description of %s: %s", name, dsc );
+            Log_dbg( LOG_DBG_FETCH, "Description of %s: %s", name, dsc );
             Grp_setDsc( name, dsc );
         }
     }
-    return TRUE;
+    Lock_closeDatabases();
+    del_DynStr( response );
+    return result;
 }
 
-Bool
+int
+Client_getDsc( void )
+{
+    GroupEnum *ge;
+    const char *pattern;
+    Bool doneOne, noServerPattern;
+    int res;
+
+    Log_inf( "Querying group descriptions" );
+
+    doneOne = FALSE;
+    res = STAT_OK;
+    ge = new_GetGrEn( client.serv );
+    while ( res == STAT_OK && ( pattern = GrEn_next( ge ) ) != NULL )
+    {
+	res = doGetDsc( pattern, &noServerPattern );
+	doneOne = TRUE;
+	if ( noServerPattern )
+	    break;
+    }
+
+    if ( ! doneOne )
+	res = doGetDsc( "", &noServerPattern );
+
+    del_GrEn( ge );
+    return res;
+}
+
+int
 Client_getNewgrps( const time_t *lastTime )
 {
     Str s;
     const char *p;
+    int stat;
 
     ASSERT( *lastTime > 0 );
-    if ( ! client.connected )
-        return FALSE;
     strftime( s, MAXCHAR, "%Y%m%d %H%M00", gmtime( lastTime ) );
     /*
       Do not use century for working with old server software until 2000.
@@ -536,20 +676,24 @@ Client_getNewgrps( const time_t *lastTime )
     */
     p = s + 2;
     if ( ! putCmd( "NEWGROUPS %s GMT", p ) )
-        return FALSE;
-    if ( getStat() != STAT_NEW_GRP_FOLLOW )
+        return STAT_CONNECTION_LOST;
+    stat = getStat();
+    if ( stat != STAT_NEW_GRP_FOLLOW )
     {
         Log_err( "NEWGROUPS command failed: %s", client.lastStat );
-        return FALSE;
+        return stat;
     }
-    processGrps();
-    return TRUE;
+
+    if( processGrps( TRUE ) == FALSE )
+	return STAT_CONNECTION_LOST;
+
+    return STAT_OK;
 }
 
 static const char *
 readField( Str result, const char *p )
 {
-    size_t len;
+    int len;
     char *r;
 
     if ( ! p )
@@ -560,7 +704,10 @@ readField( Str result, const char *p )
     while ( *p != '\t' && *p != '\n' )
     {
         if ( ! *p )
+	{
+	    *r = '\0';
             return p;
+	}
         *(r++) = *(p++);
         ++len;
         if ( len >= MAXCHAR - 1 )
@@ -576,7 +723,8 @@ readField( Str result, const char *p )
 
 static Bool
 parseOvLn( Str line, int *numb, Str subj, Str from,
-           Str date, Str msgId, Str ref, size_t *bytes, size_t *lines )
+           Str date, Str msgId, Str ref,
+	   unsigned long *bytes, unsigned long *lines )
 {
     const char *p;
     Str t;
@@ -592,10 +740,10 @@ parseOvLn( Str line, int *numb, Str subj, Str from,
     p = readField( t, p );
     *bytes = 0;
     *lines = 0;
-    if ( sscanf( t, "%d", bytes ) != 1 )
+    if ( sscanf( t, "%lu", bytes ) != 1 )
         return TRUE;
     p = readField( t, p );
-    if ( sscanf( t, "%d", lines ) != 1 )
+    if ( sscanf( t, "%lu", lines ) != 1 )
         return TRUE;
     return TRUE;
 }
@@ -623,7 +771,9 @@ nextXref( const char *pXref, Str grp, int *numb )
         return NULL;
     }
     Utl_cpyStrN( grp, s, pColon - s );
-    Log_dbg( "client.c: nextXref: grp '%s' numb %lu", grp, numb );
+    Log_dbg( LOG_DBG_FETCH,
+	     "client.c: nextXref: grp '%s' numb %lu",
+	     grp, numb );
     return src;
 }
 
@@ -632,19 +782,20 @@ needsMark( const char *ref )
 {
     Bool interesting, result;
     const char *msgId;
-    int status;
+    unsigned status;
     time_t lastAccess, nowTime;
-    double threadFollowTime, secPerDay, maxTime, timeSinceLastAccess;
+    double threadFollowTime, maxTime, timeSinceLastAccess;
     ItemList *itl;
+    const double secPerDay = 24.0 * 3600.0;
 
-    Log_dbg( "Checking references '%s' for thread mode", ref );
+    ASSERT( Lock_gotLock() );
+    Log_dbg( LOG_DBG_FETCH, "Checking references '%s' for thread mode", ref );
     result = FALSE;
     itl = new_Itl( ref, " \t" );
     nowTime = time( NULL );
     threadFollowTime = (double)Cfg_threadFollowTime();
-    secPerDay = 24.0 * 3600.0;
     maxTime = threadFollowTime * secPerDay;
-    Log_dbg( "Max time = %.0f", maxTime );
+    Log_dbg( LOG_DBG_FETCH, "Max time = %.0f", maxTime );
     for ( msgId = Itl_first( itl ); msgId != NULL; msgId = Itl_next( itl ) )
     {
         /*
@@ -657,7 +808,8 @@ needsMark( const char *ref )
             lastAccess = Db_lastAccess( msgId );
             interesting = ( status & DB_INTERESTING );
             timeSinceLastAccess = difftime( nowTime, lastAccess );
-            Log_dbg( "Msg ID '%s': since last access = %.0f, interesting = %s",
+            Log_dbg( LOG_DBG_FETCH,
+		     "Msg ID '%s': since last access = %.0f, interesting = %s",
                      msgId, timeSinceLastAccess, ( interesting ? "y" : "n" ) );
             if ( interesting && timeSinceLastAccess <= maxTime )
             {
@@ -667,11 +819,12 @@ needsMark( const char *ref )
         }
         else
         {
-            Log_dbg( "MsgID '%s': not in database.", msgId );
+            Log_dbg( LOG_DBG_FETCH, "MsgID '%s': not in database.", msgId );
         }
     }
     del_Itl( itl );
-    Log_dbg( "Article %s marking for download.",
+    Log_dbg( LOG_DBG_FETCH,
+	     "Article %s marking for download.",
              ( result ? "needs" : "doesn't need" ) );
     return result;
 }
@@ -683,13 +836,16 @@ prepareEntry( Over *ov )
     const char *msgId, *p, *xref;
     int n;
 
+    ASSERT( Lock_gotLock() );
     msgId = Ov_msgId( ov );
     if ( Pseudo_isGeneralInfo( msgId ) )
-        Log_dbg( "Skipping general info '%s'", msgId );
+        Log_dbg( LOG_DBG_FETCH, "Skipping general info '%s'", msgId );
     else if ( Db_contains( msgId ) )
     {
         xref = Db_xref( msgId );
-        Log_dbg( "Entry '%s' already in db with Xref '%s'", msgId, xref );
+        Log_dbg( LOG_DBG_FETCH,
+		 "Entry '%s' already in db with Xref '%s'",
+		 msgId, xref );
         p = nextXref( xref, g, &n );
         if ( p == NULL )
             Log_err( "Overview with no group in Xref '%s'", msgId );
@@ -698,7 +854,8 @@ prepareEntry( Over *ov )
             /* TODO: This code block seems unnessesary. Can we remove it? */
             if ( Cfg_servIsPreferential( client.serv, Grp_server( g ) ) )
             {
-                Log_dbg( "Changing first server for '%s' from '%s' to '%s'",
+                Log_dbg( LOG_DBG_FETCH,
+			 "Changing first server for '%s' from '%s' to '%s'",
                          msgId, Grp_server( g ), client.serv );
                 snprintf( t, MAXCHAR, "%s:%d %s",
                           client.grp, Ov_numb( ov ), xref );
@@ -706,7 +863,8 @@ prepareEntry( Over *ov )
             }
             else
             {
-                Log_dbg( "Adding '%s' to Xref of '%s'", g, msgId );
+                Log_dbg( LOG_DBG_FETCH,
+			 "Adding '%s' to Xref of '%s'", g, msgId );
                 snprintf( t, MAXCHAR, "%s %s:%d",
                           xref, client.grp, Ov_numb( ov ) );
                 Db_setXref( msgId, t );
@@ -715,204 +873,386 @@ prepareEntry( Over *ov )
     }
     else
     {
-        Log_dbg( "Preparing '%s' in database", msgId );
+        Log_dbg( LOG_DBG_FETCH, "Preparing '%s' in database", msgId );
         Db_prepareEntry( ov, client.grp, Ov_numb( ov ) );
     }
 }
 
-Bool
-Client_getOver( int rmtFirst, int rmtLast, FetchMode mode )
+int
+Client_getOver( const char *grp, int rmtFirst, int rmtLast, FetchMode mode )
 {
-    Bool err;
-    size_t bytes, lines;
-    int rmtNumb, oldLast, cntMarked;
+    unsigned long nbytes, nlines;
+    int rmtNumb, groupsNumb, oldLast, cntMarked;
     Over *ov;
-    Str line, subj, from, date, msgId, ref;
+    Str line, subj, from, date, msgId, ref, groups;
+    DynStr *response, *newsgroups;
+    const char *lines, *groupLines;
+    char *p;
+    FilterAction action;
+    int stat;
 
-    ASSERT( strcmp( client.grp, "" ) != 0 );
-    if ( ! client.connected )
-        return FALSE;
-    if ( ! putCmd( "XOVER %lu-%lu", rmtFirst, rmtLast ) )
-        return FALSE;
-    if ( getStat() != STAT_OVERS_FOLLOW )
+    ASSERT( ! Lock_gotLock() );
+    ASSERT( strcmp( grp, "" ) != 0 );
+
+    /* Do we need the article Newsgroups: for filtering? */
+    if ( Flt_getNewsgroups() )
     {
-        Log_err( "XOVER command failed: %s", client.lastStat );
-        return FALSE;
+	if ( ! putCmd( "XHDR Newsgroups %lu-%lu", rmtFirst, rmtLast ) )
+	    return STAT_CONNECTION_LOST;
+	stat = getStat();
+	if ( stat != STAT_HEAD_FOLLOWS )
+	{
+	    Log_err( "XHDR command failed: %s", client.lastStat );
+	    return stat;
+	}
+
+	Log_dbg( LOG_DBG_FETCH,
+		 "Requesting Newsgroups headers for remote %lu-%lu",
+		 rmtFirst, rmtLast );
+
+	newsgroups = collectTxt();
+	if ( newsgroups == NULL )
+	    return STAT_CONNECTION_LOST;
+	
+	groupLines = DynStr_str( newsgroups );
     }
-    Log_dbg( "Requesting overview for remote %lu-%lu", rmtFirst, rmtLast );
+    else
+    {
+	groupLines = NULL;
+	newsgroups = NULL;
+    }
+    
+    if ( ! putCmd( "XOVER %lu-%lu", rmtFirst, rmtLast ) )
+    {
+	del_DynStr( newsgroups );
+        return STAT_CONNECTION_LOST;
+    }
+    
+    stat = getStat();
+    if ( stat != STAT_OVERS_FOLLOW )
+    {
+	del_DynStr( newsgroups );
+        Log_err( "XOVER command failed: %s", client.lastStat );
+        return stat;
+    }
+    Log_dbg( LOG_DBG_FETCH,
+	     "Requesting overview for remote %lu-%lu",
+	     rmtFirst, rmtLast );
+
+    response = collectTxt();
+    if ( response == NULL )
+    {
+	del_DynStr( newsgroups );
+	return STAT_CONNECTION_LOST;
+    }
+
+    if ( ! Lock_openDatabases() )
+    {
+	del_DynStr( newsgroups );
+	del_DynStr( response );
+	return STAT_NEWSBASE_FATAL;
+    }
+    
+    Cont_read( grp );
     oldLast = Cont_last();
     cntMarked = 0;
-    while ( getTxtLn( line, &err ) && ! err )
+    lines = DynStr_str( response );
+    while ( ( lines = Utl_getLn( line, lines ) ) != NULL )
     {
         if ( ! parseOvLn( line, &rmtNumb, subj, from, date, msgId, ref,
-                          &bytes, &lines ) )
+                          &nbytes, &nlines ) )
             Log_err( "Bad overview line: %s", line );
+	else if ( Cont_find( msgId ) >= 0 )
+	    Log_inf( "Already have '%s'", msgId );
         else
         {
-            ov = new_Over( subj, from, date, msgId, ref, bytes, lines );
-            Cont_app( ov );
-            prepareEntry( ov );
-            if ( mode == FULL || ( mode == THREAD && needsMark( ref ) ) )
+            ov = new_Over( subj, from, date, msgId, ref, nbytes, nlines );
+	    groupsNumb = 0;
+	    p = NULL;
+	    if ( groupLines != NULL )
+	    {
+		do
+		{
+		    groupLines = Utl_getLn( groups, groupLines );
+		    groupsNumb = strtoul( groups, &p, 10 );
+		} while ( groupLines != NULL
+			  && p > groups
+			  && groupsNumb < rmtNumb );
+		if ( groupsNumb != rmtNumb )
+		    p = NULL;
+	    }
+
+	    action = Flt_checkFilters( grp, p, ov, mode );
+	    if ( action == FILTER_DISCARD )
             {
-                Req_add( client.serv, msgId );
-                ++cntMarked;
+                del_Over( ov );
             }
+	    else
+	    {
+		Cont_app( ov );     /* Cont modules owns ov after this */
+		prepareEntry( ov );
+		if ( action == FILTER_FULL
+		     || ( action == FILTER_THREAD && needsMark( ref ) ) )
+		{
+		    Req_add( client.serv, msgId );
+		    ++cntMarked;
+		}
+	    }
         }
         Grp_setRmtNext( client.grp, rmtNumb + 1 );
     }
     if ( oldLast != Cont_last() )
+    {
         Log_inf( "Added %s %lu-%lu", client.grp, oldLast + 1, Cont_last() );
-    Log_inf( "%u articles marked for download in %s", cntMarked, client.grp  );
-    return err;
+	Log_inf( "%u articles marked for download in %s",
+		 cntMarked, client.grp  );
+	if ( Cont_write() )
+            Grp_setFirstLast( grp, Cont_first(), Cont_last() );
+	Grp_setLastPostTime( grp );
+    }
+    Lock_closeDatabases();
+    del_DynStr( response );
+    del_DynStr( newsgroups );
+    return STAT_OK;
 }
 
 static void
 retrievingFailed( const char* msgId, const char *reason )
 {
-    int status;
+    unsigned status;
 
+    ASSERT( ! Lock_gotLock() );
     Log_err( "Retrieving of %s failed: %s", msgId, reason );
+    if ( ! Lock_openDatabases() )
+	return;
     status = Db_status( msgId );
     Pseudo_retrievingFailed( msgId, reason );
     Db_setStatus( msgId, status | DB_RETRIEVING_FAILED );
+    Lock_closeDatabases();
+    return;
 }
 
-static Bool
-retrieveAndStoreArt( const char *msgId )
+static int
+retrieveAndStoreArt( const char *msgId, int artcnt, int artmax )
 {
     Bool err;
     DynStr *s = NULL;
-    Str line;
 
-    Log_inf( "Retrieving %s", msgId );
-    s = new_DynStr( 5000 );
-    while ( getTxtLn( line, &err ) && ! err )
-        DynStr_appLn( s, line );
-    if ( ! err )
-        Db_storeArt( msgId, DynStr_str( s ) );
+    ASSERT( ! Lock_gotLock() );
+    Log_inf( "[%d/%d] Retrieving %s", artcnt, artmax, msgId );
+    err = TRUE;
+
+    s = collectTxt();
+    if ( s != NULL )
+    {
+	const char *txt;
+	
+	txt = DynStr_str( s );
+	if ( ! Lock_openDatabases() )
+	{
+	    del_DynStr( s );
+	    retrievingFailed( msgId, "Can't open message base" );
+	    return STAT_NEWSBASE_FATAL;
+	}
+	
+        err = ! Db_storeArt( msgId, txt );
+	if ( ! err )
+	{
+	    Str supersedeIds;
+
+	    if ( Prt_searchHeader( txt, "Supersedes", supersedeIds ) )
+	    {
+		ItemList *ids;
+		const char *supersededMsgId;
+
+		ids = new_Itl( supersedeIds, " \n\t" );
+		for ( supersededMsgId = Itl_first( ids );
+		      supersededMsgId != NULL;
+		      supersededMsgId = Itl_next( ids ) )
+		    Ctrl_cancel( supersededMsgId );
+		del_Itl( ids );
+	    }
+	}
+	Lock_closeDatabases();
+	del_DynStr( s );
+    }
     else
+    {
         retrievingFailed( msgId, "Connection broke down" );
-    del_DynStr( s );
-    return ! err;
+	return STAT_CONNECTION_LOST;
+    }
+    return err ? STAT_NEWSBASE_FATAL : STAT_OK;
 }
 
-void
+int
 Client_retrieveArt( const char *msgId )
 {
-    if ( ! client.connected )
-        return;
+    int res;
+    
+    ASSERT( Lock_gotLock() );
     if ( ! Db_contains( msgId ) )
     {
         Log_err( "Article '%s' not prepared in database. Skipping.", msgId );
-        return;
+        return STAT_PROGRAM_FAULT;
     }
     if ( ! ( Db_status( msgId ) & DB_NOT_DOWNLOADED ) )
     {
         Log_inf( "Article '%s' already retrieved. Skipping.", msgId );
-        return;
+        return STAT_OK;
     }
+
+    Lock_closeDatabases();
     if ( ! putCmd( "ARTICLE %s", msgId ) )
+    {
         retrievingFailed( msgId, "Connection broke down" );
-    else if ( getStat() != STAT_ART_FOLLOWS )
+	res = STAT_CONNECTION_LOST;
+    }
+    else if ( ( res = getStat() ) != STAT_ART_FOLLOWS )
         retrievingFailed( msgId, client.lastStat );
     else
-        retrieveAndStoreArt( msgId );
+        res = retrieveAndStoreArt( msgId, 0, 0 );
+    if ( ! Lock_openDatabases() )
+	res = STAT_NEWSBASE_FATAL;
+    return res;
 }
 
-void
-Client_retrieveArtList( const char *list )
+int
+Client_retrieveArtList( const char *list, int *artcnt, int artmax )
 {
     Str msgId;
     DynStr *s;
     const char *p;
+    int res, msgStat;
     
-    if ( ! client.connected )
-        return;
+    ASSERT( Lock_gotLock() );
     Log_inf( "Retrieving article list" );
     s = new_DynStr( (int)strlen( list ) );
     p = list;
+    res = STAT_OK;
     while ( ( p = Utl_getLn( msgId, p ) ) )
         if ( ! Db_contains( msgId ) )
-            Log_err( "Skipping retrieving of %s (not prepared in database)",
-                     msgId );
+	{
+            Log_err( "[%d/%d] Skipping retrieving of %s "
+		     "(not prepared in database)",
+                     ++(*artcnt), artmax, msgId );
+	    res = STAT_PROGRAM_FAULT;
+	}
         else if ( ! ( Db_status( msgId ) & DB_NOT_DOWNLOADED ) )
-            Log_inf( "Skipping %s (already retrieved)", msgId );
+            Log_inf( "[%d/%d] Skipping %s (already retrieved)",
+		     ++(*artcnt), artmax, msgId );
         else if ( ! putCmdNoFlush( "ARTICLE %s", msgId ) )
         {
             retrievingFailed( msgId, "Connection broke down" );
             del_DynStr( s );
-            return;
+            return STAT_CONNECTION_LOST;
         }
         else
             DynStr_appLn( s, msgId );
+
+    Lock_closeDatabases();
     fflush( client.out );
-    Log_dbg( "[S FLUSH]" );
+    Log_dbg( LOG_DBG_PROTOCOL, "[S FLUSH]" );
+    
+    /*
+     * We got something. Try to process all messages and return the
+     * 'worst' error encountered (note we may have already hit a
+     * STAT_PROGRAM_FAULT).
+     */
     p = DynStr_str( s );
-    while ( ( p = Utl_getLn( msgId, p ) ) )
+    while ( ! IS_FATAL( res ) && ( p = Utl_getLn( msgId, p ) ) )
     {
-        if ( getStat() != STAT_ART_FOLLOWS )
+	msgStat = getStat();
+	if ( msgStat == STAT_ART_FOLLOWS )
+	    msgStat = retrieveAndStoreArt( msgId, ++(*artcnt), artmax );
+	else
             retrievingFailed( msgId, client.lastStat );
-        else if ( ! retrieveAndStoreArt( msgId ) )
-            break;
+	    
+	if ( res == STAT_OK || ( ! IS_FATAL( res ) && IS_FATAL( msgStat ) ) )
+	    res = msgStat;
     }
     del_DynStr( s );
+    if ( ! Lock_openDatabases() && ! IS_FATAL( res ) )
+	res = STAT_NEWSBASE_FATAL;
+    return res;
 }
 
-Bool
+int
 Client_changeToGrp( const char* name )
 {
     unsigned int stat;
-    int estimatedNumb, first, last;
+    int estimatedNumb, first, last, res;
 
-    if ( ! client.connected )
-        return FALSE;
+    ASSERT( Lock_gotLock() );
     if ( ! Grp_exists( name ) )
-        return FALSE;
+        return STAT_NEWSBASE_FATAL;
+    Lock_closeDatabases();
+    stat = STAT_OK;
     if ( ! putCmd( "GROUP %s", name ) )
-        return FALSE;
-    if ( getStat() != STAT_GRP_SELECTED )
-        return FALSE;
+	res = STAT_CONNECTION_LOST;
+    if ( stat == STAT_OK )
+	stat = getStat();
+    if ( ! Lock_openDatabases() )
+	return STAT_NEWSBASE_FATAL;
+    if ( stat != STAT_GRP_SELECTED )
+	return stat;
     if ( sscanf( client.lastStat, "%u %d %d %d",
                  &stat, &estimatedNumb, &first, &last ) != 4 )
     {
         Log_err( "Bad server response to GROUP: %s", client.lastStat );
-        return FALSE;
+        return STAT_PROGRAM_FAULT;
     }
     Utl_cpyStr( client.grp, name );
     client.rmtFirst = first;
     client.rmtLast = last;
-    return TRUE;
+    return STAT_OK;
 }
 
 void
 Client_rmtFirstLast( int *first, int *last )
 {
+    ASSERT( Lock_gotLock() );
     *first = client.rmtFirst;
     *last = client.rmtLast;
 }
 
-Bool
-Client_postArt( const char *msgId, const char *artTxt,
-                    Str errStr )
+/**
+ * Post an article.
+ *
+ * Return status if there's a connection problem. Otherwise return
+ * STAT_OK. If there's an error in posting, put the error into
+ * errStr, and return STAT_OK. That is, the return value indicates if
+ * a proper transaction happened, and errStr indicates if that
+ * transaction contained a posting error.
+ */
+int
+Client_postArt( const char *msgId, const char *artTxt, Str errStr )
 {
-    if ( ! client.connected )
-        return FALSE;
+    int stat;
+    
+    errStr[0] = '\0';
+    
     if ( ! putCmd( "POST" ) )
-        return FALSE;
-    if ( getStat() != STAT_SEND_ART )
+        return STAT_CONNECTION_LOST;
+    stat = getStat();
+    if ( IS_FATAL( stat ) )
+	return stat;
+    else if ( stat != STAT_SEND_ART )
     {
         Log_err( "Posting of %s not allowed: %s", msgId, client.lastStat );
-        strcpy( errStr, client.lastStat );
-        return FALSE;
+        Utl_cpyStr( errStr, client.lastStat );
+        return STAT_OK;
     }
     putTxtBuf( artTxt );
     putEndOfTxt();
-    if ( getStat() != STAT_POST_OK )
+    stat = getStat();
+    if ( IS_FATAL( stat ) )
+	return stat;
+    else if ( stat != STAT_POST_OK )
     {
         Log_err( "Posting of %s failed: %s", msgId, client.lastStat );
-        strcpy( errStr, client.lastStat );
-        return FALSE;
+        Utl_cpyStr( errStr, client.lastStat );
+        return STAT_OK;
     }
     Log_inf( "Posted %s (Status: %s)", msgId, client.lastStat );
-    return TRUE;
+    return STAT_OK;
 }
