@@ -3,7 +3,7 @@
 
   Common declarations.
 
-  $Id: common.h,v 1.2 2000/05/05 23:49:38 uh1763 Exp $
+  $Id: common.h,v 1.4 2003/04/03 17:21:20 bears Exp $
 */
 
 #ifndef COMMON_H
@@ -20,6 +20,9 @@
 #define FALSE 0
 #define TRUE !0
 #define MAXCHAR 2048
+#define	MAXCHAR_STR "2048"
+/* limited sscanf() format string: */
+#define MAXCHAR_FMT "%2048s"
 
 #ifdef DEBUG
 #include <assert.h>
