@@ -8,7 +8,7 @@
   filename SPOOLDIR/overview/GROUPNAME. One entire overview file is read
   and cached in memory, at a time.
 
-  $Id: content.h,v 1.3 2000/05/06 16:56:50 bears Exp $ 
+  $Id: content.h,v 1.5 2002/06/26 13:15:44 bears Exp $ 
 */
 
 #ifndef CONT_H
@@ -22,7 +22,8 @@
 
 /*
   Try to read overviews from overview file for group <grp>.
-  Fill with fake articles, if something goes wrong.
+  Ignore any badly-formatted lines and create a new empty
+  overview file if no lines can be read.
 */
 void
 Cont_read( const char *grp );
@@ -35,8 +36,8 @@ Cont_read( const char *grp );
 void
 Cont_app( Over *ov );
 
-/* Write content */
-void
+/* Write content. Return FALSE on error. */
+Bool
 Cont_write( void );
 
 Bool
@@ -69,4 +70,7 @@ Cont_firstGrp( Str result );
 void
 Cont_expire( void );
 
+Bool
+Cont_exists( const char *grp );
+
 #endif
