@@ -1,7 +1,7 @@
 /*
   database.c
 
-  $Id: database.c,v 1.3.2.3 2001/02/28 19:03:35 enz Exp $
+  $Id: database.c,v 1.22 2003/02/13 10:04:01 bears Exp $
 
   Uses GNU gdbm library. Using Berkeley db (included in libc6) was
   cumbersome. It is based on Berkeley db 1.85, which has severe bugs
@@ -14,22 +14,25 @@
 #endif
 
 #include <stdio.h>
-#include "database.h"
 #include <ctype.h>
 #include <errno.h>
-#include <fcntl.h>
 #include <gdbm.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include "configfile.h"
-#include "itemlist.h"
+#include "content.h"
+#include "database.h"
+#include "group.h"
 #include "log.h"
 #include "protocol.h"
 #include "util.h"
-#include "wildmat.h"
 #include "portable.h"
 
+static const char ARTICLE_FILENAME_FMT[] = "%s/data/articles.gdbm";
+static const char ARTICLE_NEW_FILENAME_FMT[] = "%s/data/articles.gdbm.new";
+
+
 static struct Db
 {
     GDBM_FILE dbf;
@@ -41,7 +44,7 @@ static struct Db
     Str msgId;
 
     /* Status of loaded article */
-    int status; /* Flags */
+    unsigned status; /* Flags */
     time_t lastAccess;
 
     /* Overview of loaded article */
@@ -50,8 +53,8 @@ static struct Db
     Str date;
     Str ref;
     Str xref;
-    size_t bytes;
-    size_t lines;
+    unsigned long bytes;
+    unsigned long lines;
 
     /* Article text (except for overview header lines) */
     DynStr *txt;
@@ -61,7 +64,7 @@ static struct Db
 static const char *
 errMsg( void )
 {
-    if ( errno != 0 )
+    if ( gdbm_errno == GDBM_NO_ERROR )
         return strerror( errno );
     return gdbm_strerror( gdbm_errno );
 }
@@ -73,20 +76,21 @@ Db_open( void )
     int flags;
 
     ASSERT( db.dbf == NULL );
-    snprintf( name, MAXCHAR, "%s/data/articles.gdbm", Cfg_spoolDir() );
+    snprintf( name, MAXCHAR, ARTICLE_FILENAME_FMT, Cfg_spoolDir() );
     flags = GDBM_WRCREAT | GDBM_FAST;
 
-    if ( ! ( db.dbf = gdbm_open( name, 512, flags, 0644, NULL ) ) )
+    if ( ! ( db.dbf = gdbm_open( name, 512, flags, 0644, Log_gdbm_fatal ) ) )
     {
         Log_err( "Error opening %s for r/w (%s)", name, errMsg() );
         return FALSE;
     }
-    Log_dbg( "%s opened for r/w", name );
+    Log_dbg( LOG_DBG_NEWSBASE, "%s opened for r/w", name );
 
     if ( db.txt == NULL )
         db.txt = new_DynStr( 5000 );
 
-    gethostname( host, MAXCHAR );
+    if ( ! Utl_getFQDN( host ) )
+	Utl_cpyStr( host, "localhost.localdomain" );
     snprintf( db.xrefHost, MAXCHAR, "Xref: %s", host );
 
     return TRUE;
@@ -96,7 +100,7 @@ void
 Db_close( void )
 {
     ASSERT( db.dbf );
-    Log_dbg( "Closing database" );
+    Log_dbg( LOG_DBG_NEWSBASE, "Closing database" );
     gdbm_close( db.dbf );
     db.dbf = NULL;
     del_DynStr( db.txt );
@@ -129,7 +133,8 @@ loadArt( const char *msgId )
     dptr = val.dptr;
     if ( dptr == NULL )
     {
-        Log_dbg( "database.c loadArt: gdbm_fetch found no entry" );
+        Log_dbg( LOG_DBG_NEWSBASE,
+		 "database.c loadArt: gdbm_fetch found no entry" );
         return FALSE;
     }
     
@@ -141,29 +146,29 @@ loadArt( const char *msgId )
         return FALSE;
     }
     p = Utl_getLn( t, p );
-    if ( ! p || sscanf( t, "%lu", &db.lastAccess ) != 1 )
+    if ( ! p || sscanf( t, "%lu", (unsigned long *)&db.lastAccess ) != 1 )
     {
         Log_err( "Entry in database '%s' is corrupt (lastAccess)", msgId );
         return FALSE;
     }
-    p = Utl_getLn( db.subj, p );
-    p = Utl_getLn( db.from, p );
-    p = Utl_getLn( db.date, p );
-    p = Utl_getLn( db.ref, p );
-    p = Utl_getLn( db.xref, p );
+    p = Utl_getHeaderLn( db.subj, p );
+    p = Utl_getHeaderLn( db.from, p );
+    p = Utl_getHeaderLn( db.date, p );
+    p = Utl_getHeaderLn( db.ref, p );
+    p = Utl_getHeaderLn( db.xref, p );
     if ( ! p )
     {
         Log_err( "Entry in database '%s' is corrupt (overview)", msgId );
         return FALSE;
     }
-    p = Utl_getLn( t, p );
-    if ( ! p || sscanf( t, "%u", &db.bytes ) != 1 )
+    p = Utl_getHeaderLn( t, p );
+    if ( ! p || sscanf( t, "%lu", &db.bytes ) != 1 )
     {
         Log_err( "Entry in database '%s' is corrupt (bytes)", msgId );
         return FALSE;
     }
-    p = Utl_getLn( t, p );
-    if ( ! p || sscanf( t, "%u", &db.lines ) != 1 )
+    p = Utl_getHeaderLn( t, p );
+    if ( ! p || sscanf( t, "%lu", &db.lines ) != 1 )
     {
         Log_err( "Entry in database '%s' is corrupt (lines)", msgId );
         return FALSE;
@@ -192,9 +197,9 @@ saveArt( void )
     DynStr_appLn( s, db.date );
     DynStr_appLn( s, db.ref );
     DynStr_appLn( s, db.xref );
-    snprintf( t, MAXCHAR, "%u", db.bytes );
+    snprintf( t, MAXCHAR, "%lu", db.bytes );
     DynStr_appLn( s, t );
-    snprintf( t, MAXCHAR, "%u", db.lines );
+    snprintf( t, MAXCHAR, "%lu", db.lines );
     DynStr_appLn( s, t );
     DynStr_appDynStr( s, db.txt );
 
@@ -222,7 +227,7 @@ Db_prepareEntry( const Over *ov, const char *grp, int numb )
     ASSERT( grp );
 
     msgId = Ov_msgId( ov );
-    Log_dbg( "Preparing entry %s", msgId );
+    Log_dbg( LOG_DBG_NEWSBASE, "Preparing entry %s", msgId );
     if ( Db_contains( msgId ) )
         Log_err( "Preparing article twice: %s", msgId );
 
@@ -250,7 +255,7 @@ Db_storeArt( const char *msgId, const char *artTxt )
 
     ASSERT( db.dbf );
 
-    Log_dbg( "Store article %s", msgId );
+    Log_dbg( LOG_DBG_NEWSBASE, "Store article %s", msgId );
     if ( ! loadArt( msgId ) )
     {
         Log_err( "Cannot find info about '%s' in database", msgId );
@@ -270,6 +275,8 @@ Db_storeArt( const char *msgId, const char *artTxt )
     /* Read header */
     while ( ( artTxt = Utl_getHeaderLn( lineEx, artTxt ) ) != NULL )
     {
+	Bool continuation;
+	
         if ( lineEx[ 0 ] == '\0' )
         {
             DynStr_appLn( db.txt, lineEx );
@@ -277,7 +284,7 @@ Db_storeArt( const char *msgId, const char *artTxt )
         }
         /* Remove fields already in overview and handle x-noffle
            headers correctly in case of cascading NOFFLEs */
-        if ( Prt_getField( field, value, lineEx ) )
+        if ( Prt_getField( field, value, &continuation, lineEx ) )
         {
             if ( strcmp( field, "x-noffle-status" ) == 0 )
             {
@@ -299,8 +306,15 @@ Db_storeArt( const char *msgId, const char *artTxt )
 
     if ( artTxt == NULL )
     {
-	Log_err( "Article %s malformed: missing body", msgId );
-	return FALSE;
+	/*
+	 * This article has no body. Bereft of text it lies...
+	 *
+	 * I'm not completely sure how surprising the rest of
+	 * Noffle would find a body-less article, so substitute
+	 * an empty line.
+	 */
+	Log_inf( "Article %s malformed: missing body", msgId );
+	artTxt = "\n";
     }
 
     /* Read body */
@@ -312,7 +326,7 @@ Db_storeArt( const char *msgId, const char *artTxt )
 }
 
 void
-Db_setStatus( const char *msgId, int status )
+Db_setStatus( const char *msgId, unsigned status )
 {
     if ( loadArt( msgId ) )
     {
@@ -413,7 +427,7 @@ Db_header( const char *msgId )
     static DynStr *s = NULL;
 
     Str date, t;
-    int status;
+    unsigned status;
     const char *p;
 
     if ( s == NULL )
@@ -441,10 +455,10 @@ Db_header( const char *msgId )
     appendLongHeader( s, "Date:", db.date );
     appendLongHeader( s, "References:", db.ref );
     DynStr_app( s, "Bytes: " );
-    snprintf( t, MAXCHAR, "%u", db.bytes );
+    snprintf( t, MAXCHAR, "%lu", db.bytes );
     DynStr_appLn( s, t );
     DynStr_app( s, "Lines: " );
-    snprintf( t, MAXCHAR, "%u", db.lines );
+    snprintf( t, MAXCHAR, "%lu", db.lines );
     DynStr_appLn( s, t );
     appendLongHeader( s, db.xrefHost, db.xref );
     p = strstr( DynStr_str( db.txt ), "\n\n" );
@@ -468,7 +482,7 @@ Db_body( const char *msgId )
     return ( p + 2 );
 }
 
-int
+unsigned
 Db_status( const char *msgId )
 {
     if ( ! loadArt( msgId ) )
@@ -581,117 +595,180 @@ Db_next( const char** msgId )
     return ( cursor.dptr != NULL );
 }
 
-static int
-calcExpireDays( const char *msgId )
+void
+Db_compact( void )
 {
-    const char *xref;
-    ItemList *refs;
-    const char *ref;
-    int res;
+    ASSERT( db.dbf );
 
-    xref = Db_xref( msgId );
-    if ( xref[ 0 ] == '\0' )
-	return -1;
+    /*
+     * You'd think it would be sensible to do something like
+     *
+     * if ( gdbm_reorganize( db.dbf ) != 0 )
+     * 	 Log_err( "Error compacting article base: %s", errMsg() );
+     *
+     * here. But this just copies the database to a new one and renames,
+     * which is what Db_rebuild() does.
+     *
+     * So do nothing. We don't want expire to chew disc space.
+     */
+}
 
-    res = -1;
-    refs = new_Itl( xref, " :" );
-    for ( ref = Itl_first( refs ); ref != NULL; ref = Itl_next( refs ) )
+/*
+  Helper functions for database rebuild.
+*/
+
+static struct DbNew
+{
+    GDBM_FILE dbf;
+
+} dbNew = { NULL };
+
+
+static Bool
+newOpen( void )
+{
+    Str name;
+    int flags;
+
+    ASSERT( dbNew.dbf == NULL );
+    snprintf( name, MAXCHAR, ARTICLE_NEW_FILENAME_FMT, Cfg_spoolDir() );
+    flags = GDBM_WRCREAT | GDBM_FAST;
+
+    if ( ! ( dbNew.dbf = gdbm_open( name, 512, flags, 0644, Log_gdbm_fatal ) ) )
     {
-	Str pattern;
-	int days;
-	
-	Cfg_beginExpireEnum();
-	while ( ( days = Cfg_nextExpire( pattern ) ) != -1 )
-	    if ( Wld_match( ref, pattern )
-		 && ( ( days > res && res != 0 ) ||
-		      days == 0 ) )
-	    {
-		res = days;
-		Log_dbg ( "Custom expiry %d for %s in group %s",
-			  days, msgId, ref );
-		break;
-	    }
-	
-	Itl_next( refs );	/* Throw away group number */
+        Log_err( "Error opening %s for r/w (%s)", name, errMsg() );
+        return FALSE;
     }
+    Log_dbg( LOG_DBG_NEWSBASE, "%s opened for r/w", name );
+    return TRUE;
+}
 
-    if ( res == -1 )
-	res = Cfg_expire();
-    return res;
+static Bool
+newClose( Bool makeMain )
+{
+    Str newName;
+    
+    ASSERT( dbNew.dbf );
+    Log_dbg( LOG_DBG_NEWSBASE, "Closing new database" );
+    gdbm_close( dbNew.dbf );
+    dbNew.dbf = NULL;
+
+    snprintf( newName, MAXCHAR, ARTICLE_NEW_FILENAME_FMT, Cfg_spoolDir() );
+
+    if ( makeMain )
+    {
+	Str name;
+
+	ASSERT( db.dbf );
+	Db_close();
+	snprintf( name, MAXCHAR, ARTICLE_FILENAME_FMT, Cfg_spoolDir() );
+	if ( rename( newName, name ) != 0 )
+	{
+	    Log_err( "Rename %s to %s failed: %s",
+		     newName, name, strerror( errno ) );
+	    return FALSE;
+	}
+	Log_dbg( LOG_DBG_NEWSBASE, "Renamed %s to %s", newName, name );
+	return Db_open();
+    }
+    else
+    {
+	if ( unlink( newName ) != 0 )
+	{
+	    Log_err( "Unlink %s failed: %s", newName, strerror( errno ) );
+	    return FALSE;
+	}
+	Log_dbg( LOG_DBG_NEWSBASE, "Deleted %s", newName );
+	return TRUE;
+    }
+}
+
+static Bool
+newCopyArt( const char *msgId )
+{
+    datum key, val;
+
+    ASSERT( db.dbf );
+    ASSERT( dbNew.dbf );
+    key.dptr = (void *)msgId;
+    key.dsize = strlen( msgId ) + 1;
+
+    val = gdbm_fetch( db.dbf, key );
+    if ( val.dptr != NULL )
+    {
+	Bool res;
+	
+	res = ( gdbm_store( dbNew.dbf, key, val, GDBM_INSERT ) == 0 );
+	if ( ! res )
+	    Log_err( "Could not store %s in new database (%s)",
+		     msgId, errMsg() );
+	free( val.dptr );
+	return res;
+    }
+    Log_err( "%s not found in database", msgId );
+    return FALSE;
+}
+
+static Bool
+newContains( const char *msgId )
+{
+    datum key;
+
+    ASSERT( dbNew.dbf );
+    key.dptr = (void*)msgId;
+    key.dsize = strlen( msgId ) + 1;
+    return gdbm_exists( dbNew.dbf, key );
 }
 
 Bool
-Db_expire( void )
+Db_rebuild( void )
 {
-    int cntDel, cntLeft, flags, expDays;
-    time_t nowTime, lastAccess;
+    const Over *ov;
+    int i;
+    Str grp;
     const char *msgId;
-    Str name, tmpName;
-    GDBM_FILE tmpDbf;
-    datum key, val;
+    Bool err;
 
-    if ( ! Db_open() )
+    if ( ! Cont_firstGrp( grp ) )
         return FALSE;
-    snprintf( name, MAXCHAR, "%s/data/articles.gdbm", Cfg_spoolDir() );
-    snprintf( tmpName, MAXCHAR, "%s/data/articles.gdbm.new", Cfg_spoolDir() );
-    flags = GDBM_NEWDB | GDBM_FAST;
-    if ( ! ( tmpDbf = gdbm_open( tmpName, 512, flags, 0644, NULL ) ) )
+    if ( ! newOpen() )
+	return FALSE;
+    
+    Log_inf( "Rebuilding article database" );
+    err = FALSE;
+    do
     {
-        Log_err( "Error opening %s for read/write (%s)", tmpName, errMsg() );
-        Db_close();
-        return FALSE;
-    }
-    Log_inf( "Expiring articles" );
-    cntDel = 0;
-    cntLeft = 0;
-    nowTime = time( NULL );
-    if ( Db_first( &msgId ) )
-        do
+	if ( ! Grp_exists( grp ) )
+            Log_err( "Overview file for unknown group %s exists", grp );
+        else
         {
-	    expDays = calcExpireDays( msgId );
-            lastAccess = Db_lastAccess( msgId );
-	    if ( expDays == -1 )
-		Log_err( "Internal error: Failed expiry calculation on %s",
-			 msgId );
-	    else if ( lastAccess == -1 )
-                Log_err( "Internal error: Getting lastAccess of %s failed",
-                         msgId );
-            else if ( expDays > 0
-		      && difftime( nowTime, lastAccess ) >
-		          ( (double) expDays * 24 * 3600 ) )
-            {
-#ifdef DEBUG
-		Str last, now;
-
-		Utl_cpyStr( last, ctime( &lastAccess ) );
-		last[ strlen( last ) - 1 ] = '\0';
-		Utl_cpyStr( now, ctime( &nowTime ) );
-		last[ strlen( now ) - 1 ] = '\0';
-                Log_dbg( "Expiring %s: last access %s, time now %s",
-			 msgId, last, now );
-#endif
-                ++cntDel;
-            }
-            else
-            {
-                ++cntLeft;
-                key.dptr = (void *)msgId;
-                key.dsize = strlen( msgId ) + 1;
-
-                val = gdbm_fetch( db.dbf, key );
-                if ( val.dptr != NULL )
+            Cont_read( grp );
+            for ( i = Cont_first(); i <= Cont_last(); ++i )
+	    {
+		if ( ! Cont_validNumb( i ) )
+		    continue;
+		
+                if ( ( ov = Cont_get( i ) ) )
                 {
-                    if ( gdbm_store( tmpDbf, key, val, GDBM_INSERT ) != 0 )
-                        Log_err( "Could not store %s in new database (%s)",
-                                 errMsg() );
-                    free( val.dptr );
+                    msgId = Ov_msgId( ov );
+		    if ( msgId == NULL )
+		    {
+			err = TRUE;
+			Log_err( "Overview in %s has no msg id", grp );
+		    }
+		    else if ( ! newContains( msgId ) )
+			err |= ! newCopyArt( msgId );
                 }
-            }
+		else
+		{    
+		    err = TRUE;
+		    Log_err( "Overview %d not available in group %s", i, grp );
+		}
+	    }
         }
-        while ( Db_next( &msgId ) );
-    Log_inf( "%lu articles deleted, %lu left", cntDel, cntLeft );
-    gdbm_close( tmpDbf );
-    Db_close();
-    rename( tmpName, name );
-    return TRUE;
+    }
+    while ( Cont_nextGrp( grp ) );
+
+    return newClose( ! err );
 }
+
