@@ -3,7 +3,7 @@
 
   List of groups that are to be fetched presently.
 
-  $Id: fetchlist.h,v 1.3 2000/05/12 17:19:38 enz Exp $
+  $Id: fetchlist.h,v 1.4 2002/02/14 17:14:34 bears Exp $
 */
 
 #ifndef FETCHLIST_H
@@ -28,7 +28,7 @@ int
 Fetchlist_size( void );
 
 Bool
-Fetchlist_contains( const char *name );
+Fetchlist_contains( const char *name, FetchMode *mode );
 
 /* Get element number index. */
 Bool
