@@ -1,7 +1,7 @@
 /*
   log.c
 
-  $Id: log.c,v 1.4 2000/05/13 15:35:17 bears Exp $
+  $Id: log.c,v 1.7 2002/12/27 21:48:25 bears Exp $
 */
 
 #if HAVE_CONFIG_H
@@ -11,15 +11,19 @@
 #include <syslog.h>
 #include <stdarg.h>
 #include "common.h"
+#include "lock.h"
 #include "log.h"
 #include "portable.h"
 
 #define MAXLENGTH 240
 
-struct
+#define	DEFAULT_DBG_MASK	LOG_DBG_NONE
+
+static struct
 {
     Bool interactive;
-} log = { FALSE };
+    unsigned debugMask;
+} log = { FALSE, DEFAULT_DBG_MASK };
 
 void
 Log_init( const char *name, Bool interactive, int facility )
@@ -73,9 +77,41 @@ Log_ntc( const char *fmt, ... )
 }
 
 void
-Log_dbg( const char *fmt, ... )
+Log_dbg( unsigned subsystem, const char *fmt, ... )
 {
-#ifdef DEBUG
-    DO_LOG( LOG_DEBUG );
-#endif
+    if ( ( subsystem & log.debugMask ) != 0 ) {
+	DO_LOG( LOG_DEBUG );
+    }
 }
+
+void
+Log_setDbgMask( unsigned mask )
+{
+    /* A non-zero mask always include Noffle logs */
+    if ( mask != 0 )
+	mask |= LOG_DBG_NOFFLE;
+
+    log.debugMask = mask;
+}
+
+/*
+ * A fatal error. Log it, close down as much as possible and
+ * exit with EXIT_FAILURE.
+ */
+void
+Log_fatal( const char *fmt, ... )
+{
+    DO_LOG( LOG_ERR );
+    exit( EXIT_FAILURE );
+    /* NOTREACHED */
+}
+
+/* Fatal error function for gdbm */
+void
+Log_gdbm_fatal( const char *msg )
+{
+    Log_fatal( "gdbm: %s", msg );
+    /* NOTREACHED */
+}
+
+
