@@ -10,7 +10,7 @@
   received for some seconds (to allow multiple clients connect at the same
   time).
 
-  $Id: noffle.c,v 1.4.2.6 2000/08/11 08:17:05 enz Exp $
+  $Id: noffle.c,v 1.39 2003/05/23 09:33:10 bears Exp $
 */
 
 #if HAVE_CONFIG_H
@@ -19,39 +19,48 @@
 
 #include <ctype.h>
 #include <errno.h>
-#include <getopt.h>
 #include <signal.h>
+#include <string.h>
+#include <sys/time.h>
 #include <sys/resource.h>
 #include <syslog.h>
 #include <unistd.h>
-#include "client.h"
+#include <sys/types.h>
+#include <sys/stat.h>
 #include "common.h"
+#include "authenticate.h"
+#include "client.h"
 #include "content.h"
 #include "control.h"
 #include "configfile.h"
 #include "database.h"
+#include "expire.h"
 #include "fetch.h"
 #include "fetchlist.h"
+#include "filter.h"
 #include "group.h"
 #include "itemlist.h"
 #include "log.h"
 #include "online.h"
 #include "outgoing.h"
 #include "over.h"
+#include "post.h"
 #include "pseudo.h"
 #include "util.h"
 #include "server.h"
 #include "request.h"
 #include "lock.h"
 #include "portable.h"
-
+#include "wildmat.h"
 
 struct Noffle
 {
     Bool queryGrps;
     Bool queryDsc;
-    Bool interactive;
-} noffle = { FALSE, FALSE, TRUE };
+    Bool queryTimes;
+    Bool lockAtStartup;
+    char *serverPattern;
+} noffle = { FALSE, FALSE, FALSE, TRUE, NULL };
 
 static void
 doArt( const char *msgId )
@@ -151,24 +160,91 @@ doFetch( void )
 {
     Str serv;
 
+    if ( ! Lock_getFetchLock( LOCK_NOWAIT ) )
+    {
+	Log_err( "Another 'noffle --fetch' is in progress" );
+	return;
+    }
+    
+    Log_inf( "Fetching news..." );
     Cfg_beginServEnum();
     while ( Cfg_nextServ( serv ) )
-        if ( Fetch_init( serv ) )
-        {
-            Fetch_postArts();
+        if ( ! noffle.serverPattern 
+                || Wld_match( serv, noffle.serverPattern ) )
+            if ( Fetch_init( serv ) )
+            {
+	        Bool connOK;
+	    
+                connOK = Fetch_postArts();
 
-            Fetch_getNewGrps();
+                Flt_init( serv ); /* get filter data before processGrps() calls Utl_stamp(). */
+
+                connOK = connOK && Fetch_getNewGrps();
 
             /* Get overviews of new articles and store IDs of new articles
                that are to be fetched becase of FULL or THREAD mode in the
                request database. */
-            Fetch_updateGrps();         
+                connOK = connOK && Fetch_updateGrps();         
 
             /* get requested articles */
-            Fetch_getReq_();
+                connOK = connOK && Fetch_getReq_();
 
-            Fetch_close();
-        }
+                Fetch_close();
+            }
+
+    Log_inf( "Fetching news finished." );
+    Lock_releaseFetchLock();
+}
+
+/* List articles queued for posting to one particular server */
+static void
+listOutgoing1( const char* serv )
+{
+  Str msgid;
+
+  if ( ! Out_first( serv, msgid, NULL ) )
+      return;
+  do
+      printf( "%s %s\n", msgid, serv );
+  while ( Out_next( msgid, NULL ) );
+}
+
+/* List articles queued for posting to a particular server. */
+static void
+doOutgoing( const char *arg )
+{
+    Str serv;
+
+    Cfg_beginServEnum();   
+    while ( Cfg_nextServ( serv ) )
+	if ( ! arg || Wld_match( serv, arg ) )
+	    listOutgoing1( serv );
+}
+
+static Bool
+doPost( FILE *f, unsigned flags )
+{
+    Str line;
+    DynStr *s;
+    Bool res;
+
+    s = new_DynStr( 10000 );
+    while ( fgets( line, MAXCHAR, f ) != NULL )
+	DynStr_app( s, line );
+
+    res = TRUE;
+    if ( ! Post_open( DynStr_str( s ), flags ) )
+    {
+	fprintf( stderr, "Post failed: Malformed article.\n" );
+	res = FALSE;
+    }
+    else if ( ! Post_post() )
+    {
+	fprintf( stderr, "Post failed: Can't post to group.\n" );
+	res = FALSE;
+    }
+    Post_close();
+    return res;
 }
 
 static void
@@ -178,85 +254,31 @@ doQuery( void )
 
     Cfg_beginServEnum();
     while ( Cfg_nextServ( serv ) )
-        if ( Fetch_init( serv ) )
-        {
-            if ( noffle.queryGrps )
-                Client_getGrps();
-            if ( noffle.queryDsc )
-                Client_getDsc();
-            Fetch_close();
-        }
-}
-
-/* Expire all overviews not in database */
-static void
-expireContents( void )
-{
-    const Over *ov;
-    int i;
-    int cntDel, cntLeft;
-    Str grp;
-    Bool autoUnsubscribe;
-    int autoUnsubscribeDays;
-    time_t now = time( NULL ), maxAge = 0;
-    const char *msgId;
-
-    autoUnsubscribe = Cfg_autoUnsubscribe();
-    autoUnsubscribeDays = Cfg_autoUnsubscribeDays();
-    maxAge = Cfg_autoUnsubscribeDays() * 24 * 3600;
-    if ( ! Cont_firstGrp( grp ) )
-        return;
-    Log_inf( "Expiring overviews not in database" );
-    Fetchlist_read();
-    do
-    {
-        if ( ! Grp_exists( grp ) )
-            Log_err( "Overview file for unknown group %s exists", grp );
-        else
-        {
-            cntDel = cntLeft = 0;
-            Cont_read( grp );
-            for ( i = Cont_first(); i <= Cont_last(); ++i )
-                if ( ( ov = Cont_get( i ) ) )
-                {
-                    msgId = Ov_msgId( ov );
-                    if ( ! Db_contains( msgId ) )
-                    {
-                        Cont_delete( i );
-                        ++cntDel;
-                    }
-                    else
-                        ++cntLeft;
-                }
-            if ( ! Grp_local( grp )
-                 && Fetchlist_contains( grp )
-                 && autoUnsubscribe
-                 && difftime( now, Grp_lastAccess( grp ) ) > maxAge )
+        if ( ! noffle.serverPattern 
+                || Wld_match( serv, noffle.serverPattern ) )
+            if ( Fetch_init( serv ) )
             {
-                Log_ntc( "Auto-unsubscribing from %s after %d "
-                         "days without access",
-                         grp, autoUnsubscribeDays );
-                Pseudo_autoUnsubscribed( grp, autoUnsubscribeDays );
-                Fetchlist_remove( grp );
+	        int status = STAT_OK;
+	    
+                if ( noffle.queryGrps )
+                    status = Client_getGrps();
+                if ( status == STAT_OK && noffle.queryDsc )
+                    Client_getDsc();
+                Fetch_close();
             }
-            Cont_write();
-            Grp_setFirstLast( grp, Cont_first(), Cont_last() );
-            Log_inf( "%ld overviews deleted from group %s, %ld left (%ld-%ld)",
-                     cntDel, grp, cntLeft, Grp_first( grp ), Grp_last( grp ) );
-        }
-    }
-    while ( Cont_nextGrp( grp ) );
-    Fetchlist_write();
 }
 
 static void
 doExpire( void )
 {
-    Db_close();
-    Db_expire();
-    if ( ! Db_open() )
-        return;
-    expireContents();
+    Exp_expire();
+}
+
+static void
+doRebuild( void )
+{
+    if ( ! Db_rebuild() )
+	fprintf( stderr, "Rebuild failed.\n" );
 }
 
 static void
@@ -269,7 +291,14 @@ doCreateLocalGroup( const char * name )
     name = Utl_stripWhiteSpace( grp );
     
     if ( Grp_exists( name ) )
+    {
         fprintf( stderr, "'%s' already exists.\n", name );
+        return;
+    }
+    if ( ! Grp_isValidName( name ) )
+        fprintf( stderr, "'%s' invalid group name.\n", name );
+    else if ( Grp_isForbiddenName( name ) )
+        fprintf( stderr, "'%s' forbidden group name.\n", name );
     else
     {
         Log_inf( "Creating new local group '%s'", name );
@@ -292,7 +321,17 @@ doDeleteLocalGroup( const char * name )
     name = Utl_stripWhiteSpace( grp );
     
     if ( ! Grp_exists( name ) )
+    {
         fprintf( stderr, "'%s' does not exist.\n", name );
+        return;
+    }
+    if ( ! Grp_isValidName( name ) )
+    {
+        fprintf( stderr, "'%s' invalid group name. Skipping deletion of overviews.\n", name );
+        Log_inf( "Deleting invalid group '%s' without deleting overviews.", name );
+        Grp_delete( name );
+        printf( "Group '%s' deleted.\n", name );
+    }
     else
     {
 	int i;
@@ -300,8 +339,9 @@ doDeleteLocalGroup( const char * name )
         Log_inf( "Deleting group '%s'", name );
 
 	/*
-	  Delete all articles that are only in the group. Check the
-	  article Xref for more than one group.
+	  Delete all articles that are only in the group or are
+	  crossposted only to groups that do not exist on this
+	  server.
 	 */
 	Cont_read( name );
 	for ( i = Cont_first(); i <= Cont_last(); i++ )
@@ -310,25 +350,52 @@ doDeleteLocalGroup( const char * name )
 	    Bool toDelete;
 	    Str msgId;
 
+	    if ( ! Cont_validNumb( i ) )
+		continue;
+
 	    over = Cont_get( i );
 	    toDelete = TRUE;
 	    if ( over != NULL )
 	    {
-		ItemList * xref;
+		ItemList *xrefs;
+		const char *xref;
+		int localXrefs = 0;
 
 		Utl_cpyStr( msgId, Ov_msgId( over ) );
-		xref = new_Itl( Db_xref( msgId ), " " );
-		if ( Itl_count( xref ) > 1 )
+		xrefs = new_Itl( Db_xref( msgId ), " " );
+		for ( xref = Itl_first( xrefs );
+		      xref != NULL;
+		      xref = Itl_next( xrefs) )
+		{
+		    Str xgrp;
+		    int no;
+
+		    if ( sscanf( xref, "%"MAXCHAR_STR"[^:]:%d", xgrp, &no ) != 2 )
+		    {
+			/* Malformed xref - leave article just in case */
+			Log_err( "Malformed Xref: entry in %s: %s",
+				 msgId, xref);
+			toDelete = FALSE;
+			break;
+		    }
+		    
+		    if ( Cont_exists( xgrp ) )
+			++localXrefs;
+		}
+		
+		if ( localXrefs > 1 )
 		    toDelete = FALSE;
-		del_Itl( xref );
+		del_Itl( xrefs );
 	    }
 	    Cont_delete( i );
 	    if ( toDelete )
 		Db_delete( msgId );
 	}
-	Cont_write();
-	Grp_delete( name );
-	printf( "Group '%s' deleted.\n", name );
+        if ( Cont_write() )
+        {
+            Grp_delete( name );
+            printf( "Group '%s' deleted.\n", name );
+        }
     }
 }
 
@@ -375,7 +442,7 @@ doModify( const char *cmd, int argc, char **argv )
     else if ( strcmp( cmd, "desc" ) != 0
 	      && strcmp( cmd, "post" ) != 0 )
     {
-	fprintf( stderr, "Unknown argument -m %s\n", optarg );
+	fprintf( stderr, "Unknown argument -m %s\n", cmd );
 	return FALSE;
     }
     
@@ -411,7 +478,7 @@ doModify( const char *cmd, int argc, char **argv )
 	    Grp_setPostAllow( grp, c );
 	else
 	{
-	    fprintf( stderr, "Access must be 'y', 'n' or 'm'" );
+	    fprintf( stderr, "Access must be 'y', 'n' or 'm'\n" );
 	    return FALSE;
 	}
     }
@@ -464,7 +531,7 @@ doSubscribe( const char *name, FetchMode mode )
                 "thread" : "overview" );
     if ( ! Fetchlist_write() )
         fprintf( stderr, "Could not save fetchlist.\n" );
-    Grp_setLastAccess( name, time( NULL ) );
+    Grp_setLastAccess( name );
     return TRUE;
 }
 
@@ -476,7 +543,9 @@ doUnsubscribe( const char *name )
         printf( "%s is not in fetch list.\n", name );
     else
         printf( "%s removed from fetch list.\n", name );
-    if ( ! Fetchlist_write() )
+    if ( Fetchlist_write() )
+	Grp_setRmtNext( name, GRP_RMT_NEXT_NOT_SUBSCRIBED );
+    else
         fprintf( stderr, "Could not save fetchlist.\n" );
 }
 
@@ -487,22 +556,25 @@ printUsage( void )
       "Usage: noffle <option>\n"
       "Option is one of the following:\n"
       " -a | --article <msg id>|all      Show article(s) in database\n"
+      " -B | --rebuild                   Rebuild article database\n"
       " -c | --cancel <msg id>           Remove article from database\n"
       " -C | --create <grp>              Create a local group\n"
       " -d | --database                  Show content of article database\n"
       " -D | --delete <grp>              Delete a group\n"
       " -e | --expire                    Expire articles\n"
-      " -f | --fetch                     Get newsfeed from server/post articles\n"
+      " -f | --fetch [server]            Get newsfeed from server/post articles\n"
       " -g | --groups                    Show all groups available at server\n"
       " -h | --help                      Show this text\n"
       " -l | --list                      List groups on fetch list\n"
       " -m | --modify desc <grp> <desc>  Modify a group description\n"
-      " -m | --modify post <grp> (y|n)   Modify posting status of a local group\n"
+      " -m | --modify post <grp> (y|m|n) Modify posting status of a local group\n"
       " -n | --online                    Switch to online mode\n"
       " -o | --offline                   Switch to offline mode\n"
-      " -q | --query groups              Get group list from server\n"
-      " -q | --query desc                Get group descriptions from server\n"
-      " -r | --server                    Run as server on stdin/stdout\n"
+      " -O | --outgoing [server]         List articles queued for posting\n"
+      " -p | --post                      Post article on stdin\n"
+      " -q | --query groups [server]     Get group list from server\n"
+      " -q | --query desc [server]       Get group descriptions from server\n"
+      " -r | --server [auth]             Run as server on stdin/stdout\n"
       " -R | --requested                 List articles marked for download\n"
       " -s | --subscribe-over <grp>      Add group to fetch list (overview)\n"
       " -S | --subscribe-full <grp>      Add group to fetch list (full)\n"
@@ -512,6 +584,45 @@ printUsage( void )
     fprintf( stderr, "%s", msg );
 }
 
+    
+/* Check we are 'root' or the noffle user (usually 'news') */
+static Bool
+checkCurrentUser( void )
+{
+    if ( ! Auth_checkPrivs() )
+	return FALSE;
+    
+    /*
+     * If we're noffle.lockAtStartup, we need to drop privs now.
+     * Otherwise we're a server, and privs get dropped after authentication.
+     */
+    if ( noffle.lockAtStartup )
+	if ( ! Auth_dropPrivs() )
+	    return FALSE;
+    
+    return TRUE;
+}
+
+/*
+  Check file ownership and permissions.
+  This assumes we have cd'd to the spool directory.
+ */
+static Bool
+checkFileOwnership( void )
+{
+    Str confFile;
+    struct stat statBuf;
+    
+    Utl_cpyStr( confFile, CONFIGFILE );
+    if ( stat( confFile, &statBuf ) == 0 )
+    {
+	if ( ( statBuf.st_mode & S_IROTH ) != 0 )
+	    Log_inf( "Security warning: %s is globally readable." );
+    }
+    
+    return TRUE;
+}
+
 /*
   Allow core files: Change core limit and change working directory
   to spool directory, where news has write permissions.
@@ -532,23 +643,32 @@ enableCorefiles( void )
         Log_err( "Cannot set system core limit: %s", strerror( errno ) );
         return;
     }
-    Log_dbg( "Core limit set to %i", lim.rlim_max );
-    if ( chdir( Cfg_spoolDir() ) != 0 )
-    {
-         Log_err( "Cannot change to directory '%s'", Cfg_spoolDir() );
-         return;
-    }
-    Log_dbg( "Changed to directory '%s'", Cfg_spoolDir() );
+    Log_dbg( LOG_DBG_NOFFLE, "Core limit set to %i", lim.rlim_max );
 }
 
 static Bool
-initNoffle( Bool interactive )
+initNoffle( void )
 {
-    Log_init( "noffle", interactive, LOG_NEWS );
+    Log_init( "noffle", noffle.lockAtStartup, LOG_NEWS );
+
     Cfg_read();
-    Log_dbg( "NOFFLE version %s", Cfg_version() );
-    noffle.interactive = interactive;
-    if ( interactive )
+    Log_dbg( LOG_DBG_NOFFLE, "NOFFLE version %s", Cfg_version() );
+
+    if ( ! checkCurrentUser() )
+	return FALSE;
+    
+    /* cd to the spool directory */
+    if ( chdir( Cfg_spoolDir() ) != 0 )
+    {
+         Log_err( "Cannot change to directory '%s'", Cfg_spoolDir() );
+         return FALSE;
+    }
+    Log_dbg( LOG_DBG_NOFFLE, "Changed to directory '%s'", Cfg_spoolDir() );
+
+    if ( ! checkFileOwnership() )
+	return FALSE;
+
+    if ( noffle.lockAtStartup )
         if ( ! Lock_openDatabases() )
             return FALSE;
     enableCorefiles();
@@ -558,8 +678,9 @@ initNoffle( Bool interactive )
 static void
 closeNoffle( void )
 {
-    if ( noffle.interactive )
+    if ( Lock_gotLock() )
       Lock_closeDatabases();
+    Lock_syncDatabases();
 }
 
 static RETSIGTYPE
@@ -567,6 +688,10 @@ bugReport( int sig )
 {
     Log_err( "Received SIGSEGV. Please submit a bug report" );
     signal( SIGSEGV, SIG_DFL );
+
+    /* Attempt to save database state before passing on sig. */
+    Lock_syncDatabases();
+    
     raise( sig );
 }
 
@@ -598,37 +723,157 @@ logSignal( int sig )
     else
         Log_inf( "Received signal %i (%s). Aborting.", sig, name );
     signal( sig, SIG_DFL );
+
+    /* Attempt to save database state before passing on sig. */
+    Lock_syncDatabases();
+    
     raise( sig );
 }
 
+static void
+printInewsUsage( void )
+{
+    static const char *msg =
+	"Usage: inews [-D] [-O] [-S] [input]\n"
+	" -D    Debug - send article to stdout and don't post\n"
+	" -O    Don't add Organization header\n"
+	" -S    Don't add .signature\n"
+	" input File containing message, standard input if none specified.\n"
+	"For compatability, -h, -A, -V, -W are ignored and -N is\n"
+	"equivalent to -D.\n";
+    fprintf( stderr, "%s", msg );
+}
+
+static int
+doInews( int argc, char **argv )
+{
+    int result;
+    unsigned flags;
+    FILE *f;
+
+    UNUSED( argc );
+    
+    noffle.lockAtStartup = TRUE;
+
+    /* Process options */
+    flags = POST_ADD_ORG | POST_ADD_SIG | POST_ADD_FROM;
+    for ( ; argv[0] != NULL && argv[0][0] == '-' ; argv++ )
+    {
+	if ( argv[0][2] != '\0' )
+	{
+	    printInewsUsage();
+	    return EXIT_FAILURE;
+	}
+	
+	switch( argv[0][1] )
+	{
+	case 'h':
+	case 'A':
+	case 'V':
+	case 'W':
+	    break;
+	case 'N':
+	case 'D':
+	    flags |= POST_DEBUG;
+	    break;
+	case 'O':
+	    flags &= ~POST_ADD_ORG;
+	    break;
+	case 'S':
+	    flags &= ~POST_ADD_SIG;
+	    break;
+	default:
+	    printInewsUsage();
+	    return EXIT_FAILURE;
+	}
+    }
+    
+    if ( argv[0] ==  NULL )
+	f = stdin;
+    else
+    {
+	f = fopen( argv[0], "r" );
+	if ( f == NULL )
+	{
+	    fprintf( stderr, "Can't access %s (%s).", argv[0], strerror( errno ) );
+	    return EXIT_FAILURE;
+	}
+    }
+
+    if ( ! initNoffle() )
+	return EXIT_FAILURE;
+    result = EXIT_SUCCESS;
+
+    if ( ! doPost( f, flags ) )
+	result = EXIT_FAILURE;
+
+    if ( f != stdin )
+	fclose( f );
+
+    return result;
+}
+
+static int
+getArgLetter(const char *arg)
+{
+    int res;
+    struct option
+    {
+	const char *longOpt;
+	const char *opt;
+    } options[] =
+    {
+	{ "--article", 		"-a" },
+	{ "--cancel", 		"-c" },
+	{ "--create", 		"-C" },
+	{ "--database",		"-d" },
+	{ "--delete", 		"-D" },
+	{ "--expire", 		"-e" },
+	{ "--fetch", 		"-f" },
+	{ "--groups", 		"-g" },
+	{ "--help", 		"-h" },
+	{ "--list", 		"-l" },
+	{ "--modify", 		"-m" },
+	{ "--offline", 		"-o" },
+	{ "--online", 		"-n" },
+	{ "--outgoing", 	"-O" },
+	{ "--post", 		"-p" },
+	{ "--query", 		"-q" },
+	{ "--rebuild", 		"-B" },
+	{ "--server",		"-r" },
+	{ "--requested",	"-R" },
+	{ "--subscribe-over",	"-s" },
+	{ "--subscribe-full",	"-S" },
+	{ "--subscribe-thread",	"-t" },
+	{ "--unsubscribe",	"-u" },
+	{ "--version",		"-v" },
+	{ NULL,			NULL }
+	
+    };
+    struct option *opt;
+
+    res = -1;
+    for ( opt = options; opt->opt != NULL; opt++ )
+	if ( strcmp( arg, opt->longOpt ) == 0 ||
+	     strcmp( arg, opt->opt ) == 0 )
+	{
+	    res = opt->opt[1];
+	    break;
+	}
+
+    return res;
+}
+
+/* Options available to ordinary users, as opposed to administrators. */
+static const char* USER_OPTIONS = "adghlpRv";
+
 int main ( int argc, char **argv )
 {
     int c, result;
-    struct option longOptions[] =
-    {
-        { "article",          required_argument, NULL, 'a' },
-        { "cancel",           required_argument, NULL, 'c' },
-        { "create",           required_argument, NULL, 'C' },
-        { "database",         no_argument,       NULL, 'd' },
-        { "delete",           required_argument, NULL, 'D' },
-        { "expire",           no_argument,       NULL, 'e' },
-        { "fetch",            no_argument,       NULL, 'f' },
-        { "groups",           no_argument,       NULL, 'g' },
-        { "help",             no_argument,       NULL, 'h' },
-        { "list",             no_argument,       NULL, 'l' },
-        { "modify",           required_argument, NULL, 'm' },
-        { "offline",          no_argument,       NULL, 'o' },
-        { "online",           no_argument,       NULL, 'n' },
-        { "query",            required_argument, NULL, 'q' },
-        { "server",           no_argument,       NULL, 'r' },
-        { "requested",        no_argument,       NULL, 'R' },
-        { "subscribe-over",   required_argument, NULL, 's' },
-        { "subscribe-full",   required_argument, NULL, 'S' },
-        { "subscribe-thread", required_argument, NULL, 't' },
-        { "unsubscribe",      required_argument, NULL, 'u' },
-        { "version",          no_argument,       NULL, 'v' },
-        { NULL, 0, NULL, 0 }
-    };
+    const char *cmdname, *p;
+
+    /* Attempt to ensure databases are properly closed when we exit */
+    atexit( closeNoffle );
     
     signal( SIGSEGV, bugReport );
     signal( SIGABRT, logSignal );
@@ -637,79 +882,118 @@ int main ( int argc, char **argv )
     signal( SIGINT, logSignal );
     signal( SIGTERM, logSignal );
     signal( SIGPIPE, logSignal );
-    c = getopt_long( argc, argv, "a:c:C:dD:efghlm:onq:rRs:S:t:u:v",
-                     longOptions, NULL );
-    if ( ! initNoffle( c != 'r' ) )
+
+    /* Set umask, just in case. */
+    umask(022);
+
+    /* Find last component of command name. */
+    cmdname = argv[0];
+    p = strrchr( cmdname, '/' );
+    if ( p != NULL )
+	cmdname = ++p;
+
+    argv++;
+    argc--;
+
+    /* Were we invoked as inews? */
+    if ( strcmp( cmdname, "inews" ) == 0 )
+	return doInews( argc, argv );
+
+    c = 'h';
+    if ( *argv != NULL )
+    {
+	c = getArgLetter( *argv );
+	argv++;
+	argc--;
+    }
+
+    /*
+      If they asked for help, give it before there's a possibility
+      we could fail on init.
+     */
+    if ( c == 'h' || c == -1 )
+    {
+        printUsage();
+	return EXIT_SUCCESS;
+    }
+
+    noffle.lockAtStartup = ! ( c == 'r' );
+
+    if ( ! initNoffle() )
         return EXIT_FAILURE;
+
+    if ( ! Auth_admin() && strchr( USER_OPTIONS, c ) == NULL )
+    {
+	fprintf( stderr, "You must be a news administrator to do that.\n" );
+	return EXIT_FAILURE;
+    }
+    
     result = EXIT_SUCCESS;
     switch ( c )
     {
-    case 0:
-        /* Options that set a flag. */
-        break;
     case 'a':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -a needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            doArt( optarg );
+            doArt( *argv );
         break;
+    case 'B':
+	doRebuild();
+	break;
     case 'c':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -c needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            doCancel( optarg );
+            doCancel( *argv );
         break;
     case 'C':
-        if ( ! optarg )
+        if ( *argv == NULL  )
         {
             fprintf( stderr, "Option -C needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            doCreateLocalGroup( optarg );
+            doCreateLocalGroup( *argv );
         break;
     case 'd':
         doDb();
         break;
     case 'D':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -D needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            doDeleteLocalGroup( optarg );
+            doDeleteLocalGroup( *argv );
         break;
     case 'e':
 	doExpire();
 	break;
     case 'f':
+        noffle.serverPattern = *argv;
         doFetch();
         break;
     case 'g':
         doGrps();
         break;
-    case -1:
-    case 'h':
-        printUsage();
-        break;
     case 'l':
         doList();
         break;
     case 'm':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -m needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-	    if ( ! doModify( optarg, argc - optind, &argv[ optind ] ) )
+	    if ( ! doModify( argv[0], --argc, &argv[1] ) )
 		result = EXIT_FAILURE;
         break;
     case 'n':
@@ -718,74 +1002,99 @@ int main ( int argc, char **argv )
         else
             Online_set( TRUE );
         break;
+    case 'O':
+        doOutgoing( *argv );
+        break;
     case 'o':
         if ( ! Online_true() )
             fprintf( stderr, "NOFFLE is already offline\n" );
         else
             Online_set( FALSE );
         break;
+    case 'p':
+	if ( ! doPost( stdin, 0 ) )
+	    result = EXIT_FAILURE;
+        break;
     case 'q':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -q needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
         {
-            if ( strcmp( optarg, "groups" ) == 0 )
+            if ( strcmp( *argv, "groups" ) == 0 )
                 noffle.queryGrps = TRUE;
-            else if ( strcmp( optarg, "desc" ) == 0 )
+            else if ( strcmp( *argv, "desc" ) == 0 )
                 noffle.queryDsc = TRUE;
             else
             {
-                fprintf( stderr, "Unknown argument -q %s\n", optarg );
+                fprintf( stderr, "Unknown argument -q %s\n", *argv );
                 result = EXIT_FAILURE;
             }
+            argv++;
+            argc--;
+            noffle.serverPattern = *argv;
             doQuery();
         }
         break;
     case 'r':
-        Log_inf( "Starting as server" );
+	if ( *argv != NULL )
+	{
+	    if ( strcmp( *argv, "auth" ) != 0 )
+	    {
+                fprintf( stderr, "Unknown argument -r %s\n", *argv );
+                result = EXIT_FAILURE;
+		break;
+	    }
+
+	    Cfg_setClientAuth( TRUE );
+	    p = ", authentication enabled from command line";
+	}
+	else
+	    p = "";
+
+	Log_inf( "Starting as server%s", p );
         Server_run();
         break;
     case 'R':
-        doRequested( optarg );
+        doRequested( *argv );
         break;
     case 's':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -s needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            result = doSubscribe( optarg, OVER );
+            result = doSubscribe( *argv, OVER );
         break;
     case 'S':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -S needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            doSubscribe( optarg, FULL );
+            doSubscribe( *argv, FULL );
         break;
     case 't':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -t needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            result = doSubscribe( optarg, THREAD );
+            result = doSubscribe( *argv, THREAD );
         break;
     case 'u':
-        if ( ! optarg )
+        if ( *argv == NULL )
         {
             fprintf( stderr, "Option -u needs argument.\n" );
             result = EXIT_FAILURE;
         }
         else
-            doUnsubscribe( optarg );
+            doUnsubscribe( *argv );
         break;
     case '?':
         /* Error message already printed by getopt_long */
@@ -797,6 +1106,5 @@ int main ( int argc, char **argv )
     default:
         abort(); /* Never reached */
     }
-    closeNoffle();
     return result;
 }
