@@ -3,7 +3,7 @@
 
   Handling of pseudo articles.
 
-  $Id: pseudo.h,v 1.2 2000/05/05 23:49:38 uh1763 Exp $
+  $Id: pseudo.h,v 1.3 2000/12/05 19:50:09 enz Exp $
 */
 
 #ifndef PSEUDO_H
@@ -54,7 +54,7 @@ Pseudo_retrievingFailed( const char *msgId, const char *reason );
  */
 
 void
-Pseudo_cntInconsistent( const char *grp, int first, int last, int next );
+Pseudo_cntInconsistent( const char *grp, int first, int last, int next, int refetch );
 
 void
 Pseudo_missArts( const char *grp, int first, int next );
